/*

  MP3 Encoder demonstration.
  VLSI Solution 2012-12-20

  This VSOS application reads audio from VS1005g's default input
  (typically line input) and writes it as an MP3 file to OUT.MP3.

  This example is NOT yet demonstrative of VS1005g's quality.

  Known bugs:
  - Sometimes, depending on SD card, audio may not be continuous. This
    requires a new SD driver that is not yet available in VSOS 0.23.
  - VSOS 0.23 has a bug that may cause a linker error that says that
    SetClockSpeed is redefined. If that is the case, do the following:
    1) Open a new solution with VSOS 0.23 as the base template. If you
       already have such a solution, you can edit it.
    2) Open the file vsos02.abs.
    3) Remove the line that says: _SetClockSpeed=32764:FCN/VOID.
    4) Don't remove the other _SetClockSpeed line where the number is different. 
    5) Rebuild.
    6) Now you should be able to compile and link the MP3 encoder.

  To do:
  - User interface.

*/
#include <vo_stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sysmemory.h>
#include <encoder.h>
#include "encodermp3.h"
#include <audiofs.h>
#include <ringbuf.h>
#include <vsostasks.h>
#include <timers.h>
#include <stdbuttons.h>
#include <audio.h>
#include <lcd.h>
#include <rgb565.h>
#include "graphics.h"

/* Set any valid MP3 bitrate here. The alternatives are:
   32, 48, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320
   Note, however, that at the moment if you set bitrate to
   256 or higher, and depending on the SD card that is used,
   there may be interruptions in sound. Also note that for
   near-CD quality 160 or 192 kbit/s is recommended. */
#define BITRATE_KBPS 192

enum butActions {
  baStop = 1
};

/* The following symbols are still missing from encoder.h 2012-12-20 */
#ifndef ENCODER_QUALITY_QUALITY
#define ENCODER_QUALITY_QUALITY 0x0000U
#define ENCODER_QUALITY_VBR     0x4000U
#define ENCODER_QUALITY_ABR     0x8000U
#define ENCODER_QUALITY_CBR     0xC000U
#define ENCODER_QUALITY_MULT10  0x0000U
#define ENCODER_QUALITY_MULT100 0x1000U
#define ENCODER_QUALITY_MULT1K  0x2000U
#define ENCODER_QUALITY_MULT10K 0x3000U
#endif

__x s_int16 xmemPool[9500];
__y s_int16 ymemPool[4000];

#ifndef min
#  define min(a,b) (((a)<(b))?(a):(b))
#endif

#define RING_BUF_SIZE 12000 ///< In words, must be power of 2
#define DISK_BLOCK_SIZE 256

FILE *oFp = NULL;       ///< Output file pointer
struct Encoder *enc; ///< Encoder structure
u_int16 __align __y ringBuf[RING_BUF_SIZE];
volatile u_int16 __y *ringRP = ringBuf, *ringWP = ringBuf;
u_int16 diskBlock[DISK_BLOCK_SIZE];

auto s_int16 RingBufFill(void) {
  s_int16 res = ringWP - ringRP;
  if (res < 0)
    return res + RING_BUF_SIZE;
  return res;
}

auto s_int16 RingBufFree(void) {
  s_int16 fr = RING_BUF_SIZE - RingBufFill();
  if (fr) { // May not fill last element, full and empty buffer would look same
    fr--;
  }
  return fr;
}


u_int16 vuMeterMemory[2];
u_int16 clearVuMeter = 1;


auto void VuMeter(register __i2 const s_int16 *d, register __d0 u_int16 n,
                  register __i1 u_int16 *mem) {
  u_int16 memCopy = *mem;
  u_int16 i;
  for (i=0; i<n; i++) {
    u_int16 t = abs(*d++);
    if (t > memCopy) {
      memCopy = t;
    }
  }
  *mem = memCopy;
}


/**
   struct EncoderServices service ESRead(). Reads bytes into an u_int16
   table.

   \param es Encoder Service structure.
   \param data Pointer to destination data pointers, one for each channel.
   \param samples How many n-channel 16-bit samples to read.
        word is only half filled, only the MSB is changed.
   \param chan Number of channels.
   \return Number of samples read. 0 is an EOF or error.
*/
u_int16 ESRead(struct EncoderServices *es, s_int16 **data, u_int16 samples,
               u_int16 chan) {
  u_int16 __y i;

  if (chan == 1) {
    /* Mono */
    fread(data[0], sizeof(s_int16), samples, stdaudioin);
    fwrite(data[0], sizeof(s_int16), samples, stdaudioout);
    if (clearVuMeter) {
      vuMeterMemory[0] = vuMeterMemory[1] = 0;
    }
    VuMeter(data[0], samples, vuMeterMemory);
    data[0] += samples;
  } else {
#define STB 64
    u_int16 left = samples;
    static s_int16 stb[STB];
    while (left) {
      int toRead = min(left, STB/2);
      fread(stb, sizeof(s_int16), toRead*2, stdaudioin);
      fwrite(stb, sizeof(s_int16), toRead*2, stdaudioout);
      ringcpyXX((u_int16 *)data[0], 1, (u_int16 *)stb  , 2, toRead);
      ringcpyXX((u_int16 *)data[1], 1, (u_int16 *)stb+1, 2, toRead);
      if (clearVuMeter) {
        vuMeterMemory[0] = vuMeterMemory[1] = 0;
      }
      VuMeter(data[0], toRead, vuMeterMemory  );
      VuMeter(data[1], toRead, vuMeterMemory+1);
      data[0] += toRead;
      data[1] += toRead;
      left -= toRead;
    }
  }

  return samples;
}




/**
   struct EncoderServices service ESOutput(). Writes data to MP3 file.

   \param es Encoder Service structure.
   \param data Interleaved audio data. This data buffer may
        be used as working space by Output().
   \param bytes How many bytes of data to write.
*/
s_int16 ESOutput(struct EncoderServices *es, u_int16 *data, s_int16 bytes) {
  u_int16 words = (bytes+1)/2;

  while (RingBufFree() <= words) { // If there is buffer overflow, wait.
    Delay(5);                      // Waiting will cause interruption in sound
  }                                // but no bitstream error.

  // First write how many bytes there are to follow in the buffer.
  // We need to do this because our storage ring buffer is 16-bit memory,
  // but our data is in 8-bit (byte) quantums.
  ringWP = RING_XY_D(ringcpyXY(ringWP, MAKEMODF(RING_BUF_SIZE),
                               (u_int16 *)(&bytes), 0, 1));

  // Then write the actual data.
  ringWP = RING_XY_D(ringcpyXY(ringWP, MAKEMODF(RING_BUF_SIZE),
                               data, 1, words));

  return 0;
}



const u_int16 linToDBTab[5] = {36781, 41285, 46341, 52016, 58386};

/*
  Converts a linear 16-bit value between 0..65535 to decibels.
    Reference level: 32768 = 96dB (32767 = 95dB).
  Bugs:
    - For the input of 0, 0 dB is returned, because minus infinity cannot
      be represented with integers.
    - Assumes a ratio of 2 is 6 dB, when it actually is approx. 6.02 dB.
*/
auto u_int16 LinToDB(u_int16 n) {
  int res = 96, i;

  if (!n)               /* No signal should return minus infinity */
    return 0;

  while (n < 32768U) {  /* Amplify weak signals */
    res -= 6;
    n <<= 1;
  }

  for (i=0; i<5; i++)   /* Find exact scale */
    if (n >= linToDBTab[i])
      res++;

  return res;
}


#if 0
const u_int16 DBToLinTab[6] = {
#if 0
  /* Correct rounding, so perfect complement to LinToDB() */
  34716, 38968, 43740, 49097, 55109, 61858
#else
  /* Gives nicer powers of two but does not exactly complement LinToDB() */
  32768, 36781, 41285, 46341, 52016, 58386
#endif
};

auto u_int16 DBToLin(u_int16 n) {
  u_int16 res = 1;
  if (n < 6) {
    return 0;
  }
  while ((n-=6) >= 6) {
    res <<= 1;
  }
  res = ((u_int32)res * DBToLinTab[n] + (1<<14)) >> 15;
  return res;
}
#endif





enum EncoderError encErrCode = eeOk;
const char *encErrStr = "";

/*
  This is the encoder task. Because it "only" calls the MP3 Encoder
  function, it seems very trivial. However, this tasks does all the
  encoding, and the encoder also calls ESRead() and ESOutput()
  functions through the EncoderServices structure in enc.
*/
void EncoderTask(void) {
  encErrCode = enc->Encode((struct Encoder *)enc, &encErrStr);
}



struct EncoderServices es = {0};

StdButton buttons[4] = {0};

int main(void) {
  // Remember to never allocate buffers from stack space. So, if you
  // allocate the space inside your function, never forget "static"!
  int i;
  int retVal = EXIT_SUCCESS;
  u_int32 lastUITimeCount = ReadTimeCount();
  u_int32 lastScreenUpdateTimeCount = lastUITimeCount;
  u_int32 lastSeconds = 0;
  StdButton *textArea, *currentButton = buttons;

  /* Prepare screen */
  console.Ioctl(&console, IOCTL_START_FRAME, "MP3 Encoder Demo");
  SetVirtualResolution(12,9);

  CreateStdButton(currentButton++, baStop, BTN_NORMAL, 10,7, 2,2, "Stop");
  textArea=currentButton;
  CreateStdButton(currentButton++, BTN_END, BTN_INVISIBLE|BTN_TEXT,
                  0,0, 12,7, "");

  SetClippingRectangleToButton(textArea);

  RenderStdButtons(buttons);

  /* Prepare EncoderServices structure */
  es.Read = ESRead;
  es.Output = ESOutput;

  /* Initialize memory allocation pools for the decoder. */
  __InitMemAlloc(xmemPool, sizeof(xmemPool), ymemPool, sizeof(ymemPool));


  printf("VSOS 0.23 MP3 Encoder v0.001\n"
         "2012 VLSI Solution\n\n"
         "Note:\n"
         "This is an EARLY demonstration version. Its\n"
         "quality is NOT yet representative of\n"
         "VS1005g's true capabilities. Also the User\n"
         "Interface will change a lot in the future.\n\n");

  printf("Old input buffer size %d words\n",
         stdaudioin->op->Ioctl(stdaudioin,
                               IOCTL_AUDIO_GET_INPUT_BUFFER_SIZE, NULL));
  printf("Change input buffer size: %d (0 = ok)\n",
         stdaudioin->op->Ioctl(stdaudioin,
                               IOCTL_AUDIO_SET_INPUT_BUFFER_SIZE,
                               (void *)2048));


  oFp = fopen("S:OUT.MP3", "wb");
  if (!oFp) {
    printf("Couldn't open write file\n");
    retVal = EXIT_FAILURE;
    goto end;
  }


  /* Create the encoder */
  enc = EncMp3Create(&es, 2, 48000,
           ENCODER_QUALITY_CBR | ENCODER_QUALITY_MULT1K | BITRATE_KBPS);
  if (!enc) {
    printf("Couldn't create encoder\n");
    retVal = EXIT_FAILURE;
    goto end;
  }


  /* Start the encoder task function EncoderTask(). TASK_DECODER
     starts as a default at a higher priority than this main task. */
  StartTask(TASK_DECODER, EncoderTask);


  printf("Recording %d kbit/s MP3 file\n", BITRATE_KBPS);

  LcdFilledRectangle(17, 147, 22+160+2, 171, NULL,
                     __RGB565RGB(0, 128, 0));
  LcdFilledRectangle(20+(84-56)*4, 147, 22+160+2, 171, NULL,
                     __RGB565RGB(128, 128, 0));
  LcdFilledRectangle(20+(90-56)*4, 147, 22+160+2, 171, NULL,
                     __RGB565RGB(128, 0, 0));


  /* Main loop runs as long as EncoderTask() has not ended.
     The main loop will take care of writing the MP3 data EncoderTask()
     has created into the ring buffer. Also the user interface is handled
     here. */

  while (pSysTasks[TASK_DECODER].task.tc_State &&
         pSysTasks[TASK_DECODER].task.tc_State != TS_REMOVED) {
    u_int32 tmpTimeCount = ReadTimeCount();

    /* If there is data to write, and less then 1/10 seconds since user
       interface was last handled... */
    if (RingBufFill() > 1 && tmpTimeCount-lastUITimeCount < TICKS_PER_SEC/10) {
      u_int16 bytesLeft;
      ringRP = RING_YX_S(ringcpyYX(&bytesLeft, 0,
                                   ringRP, MAKEMODF(RING_BUF_SIZE), 1));
      while (bytesLeft) {
        u_int16 bytes = min(bytesLeft, DISK_BLOCK_SIZE);
        bytesLeft -= bytes;
        ringRP = RING_YX_S(ringcpyYX(&diskBlock, 1,
                                     ringRP, MAKEMODF(RING_BUF_SIZE),
                                     (bytes+1)/2));
        // Cannot call fwrite() because it cannot write at byte boundaries.
        // So we'll call the low-lever write function directly.
        oFp->op->Write(oFp, diskBlock, 0, bytes);
      }
    } else { /* !(RingBufFill() > 1 && ..) */
      u_int16 buttonPress = GetStdButtonPress(buttons);
      switch (buttonPress) {
      case baStop:
        //        printf("Stop pushed\n");
        es.cancel = 1; // Asks the encoder to stop.
        break;
      }

      if (es.playTimeSeconds != lastSeconds) {
        double br = es.avgBitRate*0.001;
        lcd0.x = lcd0.clipx1;
        lcd0.y = lcd0.clipy1+112;
        printf("%5lu:%02lu",
               lastSeconds/60, lastSeconds%60);
        printf(" %5.1lf kbit/s\n", es.avgBitRate*0.001);
        lastSeconds = es.playTimeSeconds;
      }

      if (tmpTimeCount-lastScreenUpdateTimeCount >= TICKS_PER_SEC/50) {
        s_int16 dB[2];
        static s_int16 oldT[2] = {0, 0};
        lastScreenUpdateTimeCount = tmpTimeCount;

        dB[0] = LinToDB(vuMeterMemory[0]+1);
        if (enc->channels == 1) {
          dB[1] = dB[0];
        } else {
          dB[1] = LinToDB(vuMeterMemory[1]+1);
        }
        clearVuMeter = 1;
        for (i=0; i<2; i++) {
          u_int16 color = __RGB565RGB(0, 255, 0);
          s_int16 t;
          
          t = (dB[i]-56)*4;
          if (t < 0) {
            t = 0;
          }
          if ((oldT[i]-=2) < 0) {
            oldT[i] = 0;
          }
          if (oldT[i] < t) {
            oldT[i] = t;
          }
          t = oldT[i];
          if (t > (90-56)*4) {
            u_int16 tt = 255-(t-((90-56)*4))*10;
            color = __RGB565RGB(255, tt, 0);
          } else if (t > (84-56)*4) {
            u_int16 tt = (t-((84-56)*4))*10;
            color = __RGB565RGB(tt, 240, 0);
          }

          //                    printf("dB[%d] = %d = %d\n", i, dB[i], t);
          LcdFilledRectangle(20, 150+i*11, 20+t, 157+i*11, NULL, color);
          if (t < 160) {
            LcdFilledRectangle(21+t, 150+i*11, 20+160, 157+i*11, NULL, 0);
          }
        } /* for (i=0; i<2; i++) */
      } /* if (over 1/50s since last updated display) */
      lastUITimeCount = tmpTimeCount;
    }  /* !(RingBufFill() > 1 && ..) */
  } /* while (encoder task running) */

  lcd0.x = lcd0.clipx1;
  lcd0.y = lcd0.clipy2-24;


 end:
  if (enc) {
    if (encErrCode > 0) {
      printf("Encode errCode %d, errStr \"%s\"\n", encErrCode, encErrStr);
      retVal = EXIT_FAILURE;
    }

    //    printf("Delete encoder\n");
    enc->Delete((struct Encoder *)enc);
    enc = NULL;
  }

  if (oFp) {
    //    printf("Close output file\n");
    fclose(oFp);
    oFp = NULL;
  }

  //  printf("End.\n");

  return retVal;
}
